from hirefire_resource.middleware.asgi import RequestInfo, request


class HireFireMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        response = await request(self.extract_request_info(scope))

        if response:
            await self.send_response(send, response)
        else:
            await self.app(scope, receive, send)

    @staticmethod
    def extract_request_info(scope):
        path = scope.get("path", "")
        headers = dict(scope.get("headers", []))
        request_start = headers.get(b"x-request-start")
        request_start_time = request_start.decode() if request_start else None
        hirefire_token = headers.get(b"hirefire-token")
        hirefire_token = hirefire_token.decode() if hirefire_token else None
        return RequestInfo(path, request_start_time, hirefire_token)

    @staticmethod
    async def send_response(send, response):
        status, headers, body = response

        headers = [(k.encode("utf-8"), v.encode("utf-8")) for k, v in headers.items()]

        await send(
            {
                "type": "http.response.start",
                "status": status,
                "headers": headers,
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": body.encode("utf-8"),
            }
        )



from hirefire_resource.middleware.asgi import RequestInfo, request


class HireFireMiddleware:
    def __init__(self, inner):
        self.inner = inner

    async def __call__(self, scope, receive, send):
        if scope["type"] == "http":
            token = None
            for header_name, header_value in scope["headers"]:
                if header_name.lower() == b"hirefire-token":
                    token = header_value.decode("utf-8")
                    break

            response = await request(
                RequestInfo(
                    path=scope["path"],
                    request_start_time=self.extract_request_start_time(scope),
                    token=token,
                )
            )

            if response:
                await self.send_response(send, response)
                return

        await self.inner(scope, receive, send)

    async def send_response(self, send, response_data):
        status, headers, body = response_data
        response_headers = [
            (key.encode("utf-8"), value.encode("utf-8"))
            for key, value in headers.items()
        ]

        await send(
            {
                "type": "http.response.start",
                "status": status,
                "headers": response_headers,
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": body.encode("utf-8"),
            }
        )

    def extract_request_start_time(self, scope):
        for header_name, header_value in scope["headers"]:
            if header_name.lower() == b"x-request-start":
                timestamp_str = header_value.decode("utf-8")
                try:
                    return int(timestamp_str)
                except ValueError:
                    pass
        return None



import logging
import sys

from hirefire_resource.web import Web
from hirefire_resource.worker import Worker


class Configuration:
    def __init__(self):
        self.web = None
        self.workers = []
        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger("hirefire_resource")
        logger.setLevel(logging.INFO)

        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)

        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)

        logger.addHandler(handler)

        return logger

    def dyno(self, name, proc=None):
        if name == "web":
            self.web = Web(self)
        else:
            self.workers.append(Worker(name, proc))



import asyncio
import json

from hirefire_resource import HireFire
from hirefire_resource.middleware import (  # noqa
    RequestInfo,
    matches_hirefire_path,
    matches_info_path,
    process_request_queue_time,
)
from hirefire_resource.version import VERSION


async def request(request_info):
    process_request_queue_time(request_info)

    if matches_hirefire_path(request_info) or matches_info_path(request_info):
        return await construct_info_response()


async def construct_info_response():
    headers = {
        "Content-Type": "application/json",
        "Cache-Control": "must-revalidate, private, max-age=0",
        "HireFire-Resource": f"Python-{VERSION}",
    }
    workers_info = await collect_workers_data()
    body = json.dumps(workers_info)

    return 200, headers, body


async def collect_workers_data():
    data = []

    for worker in HireFire.configuration.workers:
        value = worker.value()
        if asyncio.iscoroutine(value):
            value = await value
        data.append({"name": worker.name, "value": value})

    return data



import asyncio
import functools
import os
import time
from datetime import datetime

import redis


def job_queue_latency(*queues, redis_url=None):
    """
    Calculates the maximum job queue latency using RQ. If no queues are specified, it measures
    latency across all available queues.

    This function dynamically selects the Redis broker based on the provided redis_url or
    environment variables, or falls back to a default local Redis URL.

    Args:
        *queues (str): Names of the queues for latency measurement.
        redis_url (str, optional): The Redis URL. Defaults in the following order:
            - Passed argument `redis_url`.
            - Environment variables `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`, `OPENREDIS_URL`.
            - "redis://localhost:6379/0".

    Returns:
        float: The maximum latency in seconds across the specified queues.

    Examples:
        >>> job_queue_latency()
        22.918
        >>> job_queue_latency("default")
        10.172
        >>> job_queue_latency("default", "mailer")
        22.918
        >>> job_queue_latency("default", redis_url="redis://localhost:6379/0")
        10.172
    """
    redis_url = (
        redis_url
        or os.getenv("REDIS_TLS_URL")
        or os.getenv("REDIS_URL")
        or os.getenv("REDISTOGO_URL")
        or os.getenv("REDISCLOUD_URL")
        or os.getenv("OPENREDIS_URL")
        or "redis://localhost:6379/0"
    )

    redis_client = redis.Redis.from_url(redis_url)

    if not queues:
        keys = redis_client.keys("rq:scheduled:*") + redis_client.keys("rq:queue:*")
        queues = set(key.decode("utf-8").split(":")[2] for key in keys)

    pipeline = redis_client.pipeline()
    current_time = time.time()

    for queue in queues:
        pipeline.lindex(f"rq:queue:{queue}", 0)
        pipeline.zrangebyscore(
            f"rq:scheduled:{queue}",
            "-inf",
            current_time,
            withscores=True,
            start=0,
            num=1,
        )

    job_ids = pipeline.execute()

    for job_id in job_ids[::2]:
        if job_id:
            pipeline.hget(f"rq:job:{job_id.decode('utf-8')}", "enqueued_at")

    enqueued_at_times = pipeline.execute()

    max_latency = 0.0

    for enqueued_at in enqueued_at_times:
        if enqueued_at:
            latency = current_time - _iso_to_unix(enqueued_at.decode("utf-8"))
            max_latency = max(max_latency, latency)

    for job_data in job_ids[1::2]:
        if job_data:
            job_id, score = job_data[0]
            if score < current_time:
                latency = current_time - score
                max_latency = max(max_latency, latency)

    return max_latency


async def async_job_queue_latency(*queues, redis_url=None):
    """
    Asynchronously calculates the maximum job queue latency using RQ. If no queues are specified, it
    measures latency across all available queues.

    This function is an asynchronous wrapper around the synchronous `job_queue_latency` function. It
    executes the synchronous function in a separate thread using asyncio's event loop and
    `run_in_executor` method. This ensures that the synchronous Redis I/O operations do not block
    the asyncio event loop.

    Args:
        *queues (str): Names of the queues for latency measurement.
        redis_url (str, optional): The Redis URL. Defaults in the following order:
            - Passed argument `redis_url`.
            - Environment variables `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`, `OPENREDIS_URL`.
            - "redis://localhost:6379/0".

    Returns:
        float: The maximum latency in seconds across the specified queues.

    Examples:
        >>> await async_job_queue_latency()
        22.918
        >>> await async_job_queue_latency("default")
        10.172
        >>> await async_job_queue_latency("default", "mailer")
        22.918
        >>> await async_job_queue_latency("default", redis_url="redis://localhost:6379/0")
        10.172
    """
    loop = asyncio.get_event_loop()
    func = functools.partial(job_queue_latency, *queues, redis_url=redis_url)
    return await loop.run_in_executor(None, func)


def job_queue_size(*queues, redis_url=None):
    """
    Calculates the maximum job queue size using RQ. If no queues are specified, it measures latency
    across all available queues.

    This function dynamically selects the Redis broker based on the provided redis_url, environment
    variables, or falls back to a default local Redis URL.

    Args:
        *queues (str): Names of the queues for size measurement.
        redis_url (str, optional): The Redis URL. Defaults in the following order:
            - Passed argument `redis_url`.
            - Environment variables `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`, `OPENREDIS_URL`.
            - "redis://localhost:6379/0".

    Returns:
        int: The cumulative job queue size across the specified queues.

    Examples:
        >>> job_queue_size()
        127
        >>> job_queue_size("default")
        42
        >>> job_queue_size("default", "mailer")
        127
        >>> job_queue_size("default", redis_url="redis://localhost:6379/0")
        42
    """
    redis_url = (
        redis_url
        or os.getenv("REDIS_TLS_URL")
        or os.getenv("REDIS_URL")
        or os.getenv("REDISTOGO_URL")
        or os.getenv("REDISCLOUD_URL")
        or os.getenv("OPENREDIS_URL")
        or "redis://localhost:6379/0"
    )

    redis_client = redis.Redis.from_url(redis_url)

    if not queues:
        keys = redis_client.keys("rq:scheduled:*") + redis_client.keys("rq:queue:*")
        queues = set(key.decode("utf-8").split(":")[2] for key in keys)

    pipeline = redis_client.pipeline()
    current_time = int(time.time())

    for queue in queues:
        pipeline.llen(f"rq:queue:{queue}")
        pipeline.zcount(f"rq:scheduled:{queue}", 0, current_time)

    job_counts = pipeline.execute()
    total_jobs = sum(job_counts)

    return total_jobs


async def async_job_queue_size(*queues, redis_url=None):
    """
    Asynchronously calculates the maximum job queue size using RQ. If no queues are specified, it
    measures latency across all available queues.

    This function is an asynchronous wrapper around the synchronous `job_queue_size` function. It
    executes the synchronous function in a separate thread using asyncio's event loop and
    `run_in_executor` method. This ensures that the synchronous Redis I/O operations do not block
    the asyncio event loop.

    Args:
        *queues (str): Names of the queues for size measurement.
        redis_url (str, optional): The Redis URL. Defaults in the following order:
            - Passed argument `redis_url`.
            - Environment variables `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`, `OPENREDIS_URL`.
            - "redis://localhost:6379/0".

    Returns:
        int: The cumulative job queue size across the specified queues.

    Examples:
        >>> await async_job_queue_size()
        127
        >>> await async_job_queue_size("default")
        42
        >>> await async_job_queue_size("default", "mailer")
        127
        >>> await async_job_queue_size("default", redis_url ="redis://localhost:6379/0")
        42
    """
    loop = asyncio.get_event_loop()
    func = functools.partial(job_queue_size, *queues, redis_url=redis_url)
    return await loop.run_in_executor(None, func)


def _iso_to_unix(iso_time):
    dt = datetime.fromisoformat(iso_time.replace("Z", "+00:00"))
    unix_time = float(dt.timestamp())

    return unix_time



from flask import Response

from hirefire_resource.middleware.wsgi import RequestInfo, request


class HireFireMiddleware:
    def __init__(self, app):
        self.app = app
        self.original_wsgi_app = app.wsgi_app

    def __call__(self, environ, start_response):
        with self.app.request_context(environ):
            response = request(
                RequestInfo(
                    path=environ.get("PATH_INFO"),
                    request_start_time=environ.get("HTTP_X_REQUEST_START"),
                    token=environ.get("HTTP_HIREFIRE_TOKEN"),
                )
            )

            if response:
                status, headers, body = response
                response = Response(body, status=status, headers=headers)
                return response(environ, start_response)

        return self.original_wsgi_app(environ, start_response)



import http.client
import json
import os
import re
import socket
import threading
import time
from datetime import datetime

from hirefire_resource.version import VERSION


class DispatchError(Exception):
    pass


class Web:
    def __init__(self, configuration):
        self._buffer = {}
        self._mutex = threading.Lock()
        self._dispatcher_running = False
        self._dispatcher = None
        self._dispatch_interval = 1
        self._dispatch_timeout = 5
        self._buffer_ttl = 60
        self._configuration = configuration

    def start_dispatcher(self):
        with self._mutex:
            if self._dispatcher_running:
                return False
            self._dispatcher_running = True

        self._logger.info("[HireFire] Starting web metrics dispatcher.")
        self._dispatcher = threading.Thread(target=self._start_dispatcher)
        self._dispatcher.start()
        return True

    def stop_dispatcher(self):
        with self._mutex:
            if not self._dispatcher_running:
                return False
            self._dispatcher_running = False

        if self._dispatcher:
            self._dispatcher.join(self._dispatch_timeout)
            self._dispatcher = None

        self._flush_buffer()
        self._logger.info("[HireFire] Web metrics dispatcher stopped.")
        return True

    def dispatcher_running(self):
        with self._mutex:
            return self._dispatcher_running

    def add_to_buffer(self, request_queue_time):
        with self._mutex:
            timestamp = int(datetime.now().timestamp())
            self._buffer.setdefault(timestamp, []).append(request_queue_time)

    def _flush_buffer(self):
        with self._mutex:
            buffer = self._buffer
            self._buffer = {}
            return buffer

    def _dispatch_buffer(self):
        buffer = self._flush_buffer()

        if buffer:
            try:
                if os.environ.get("HIREFIRE_VERBOSE"):
                    self._logger.info(f"[HireFire] Dispatching web metrics: {buffer}")
                self._submit_buffer(buffer)
            except Exception as e:
                self._repopulate_buffer(buffer)
                self._logger.error(
                    f"[HireFire] Error while dispatching web metrics: {str(e)}"
                )

    def _start_dispatcher(self):
        while self.dispatcher_running():
            self._dispatch_buffer()
            time.sleep(self._dispatch_interval)

    def _repopulate_buffer(self, buffer):
        now = int(datetime.now().timestamp())
        with self._mutex:
            for timestamp, request_queue_times in buffer.items():
                if timestamp >= now - self._buffer_ttl:
                    self._buffer.setdefault(timestamp, []).extend(request_queue_times)

    def _submit_buffer(self, buffer):
        hirefire_token = os.environ.get("HIREFIRE_TOKEN")

        if not hirefire_token:
            raise DispatchError(
                "The HIREFIRE_TOKEN environment variable is not set. Unable to submit "
                "Request Queue Time metric data. The HIREFIRE_TOKEN can be found in "
                "the HireFire Web UI in the web dyno manager settings."
            )

        buffer_string = json.dumps(buffer)

        headers = {
            "Content-Type": "application/json",
            "HireFire-Token": hirefire_token,
            "HireFire-Resource": f"Python-{VERSION}",
        }

        hirefire_dispatch_url = re.sub(
            r"^https?://",
            "",
            os.environ.get("HIREFIRE_DISPATCH_URL", "logdrain.hirefire.io"),
        )
        connection = http.client.HTTPSConnection(
            hirefire_dispatch_url, timeout=self._dispatch_timeout
        )

        try:
            connection.request("POST", "/", buffer_string, headers)
            response = connection.getresponse()

            if response.status >= 400:
                raise DispatchError(
                    f"HTTP error occurred: {response.status} {response.reason}"
                )

            self._adjust_parameters(response)
            return response
        except http.client.HTTPException as e:
            raise DispatchError(f"HTTP error occurred: {str(e)}")
        except socket.timeout:
            raise DispatchError("The request to the server timed out.")
        except Exception as e:
            raise DispatchError(f"Error occurred during request: {str(e)}")
        finally:
            connection.close()

    def _adjust_parameters(self, response):
        if "HireFire-Resource-Dispatch-Interval" in response.headers:
            self._dispatch_interval = int(
                response.headers["HireFire-Resource-Dispatch-Interval"]
            )
        if "HireFire-Resource-Dispatch-Timeout" in response.headers:
            self._dispatch_timeout = int(
                response.headers["HireFire-Resource-Dispatch-Timeout"]
            )
        if "HireFire-Resource-Buffer-TTL" in response.headers:
            self._buffer_ttl = int(response.headers["HireFire-Resource-Buffer-TTL"])

    @property
    def _logger(self):
        return self._configuration.logger



from importlib.metadata import PackageNotFoundError, metadata

try:
    _metadata = metadata("hirefire-resource")

    __version__ = _metadata.get("Version")
    __author__ = _metadata.get("Author")
    __contact__ = _metadata.get("Author-email")
    __homepage__ = _metadata.get("Home-page")
    __keywords__ = _metadata.get("Keywords", "").split(", ")
    __docformat__ = "google"
except PackageNotFoundError:
    __version__ = "unknown"
    __author__ = "unknown"
    __contact__ = "unknown"
    __homepage__ = "unknown"
    __keywords__ = "unknown"
    __docformat__ = "unknown"

from hirefire_resource.hirefire import HireFire  # noqa



from django.http import HttpResponse

from hirefire_resource.middleware.wsgi import RequestInfo, request


class HireFireMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, req):
        response = request(
            RequestInfo(
                path=req.path,
                request_start_time=req.META.get("HTTP_X_REQUEST_START"),
                token=req.META.get("HTTP_HIREFIRE_TOKEN"),
            )
        )

        if response:
            status, headers, body = response
            response = HttpResponse(content=body, status=status, headers=headers)
            return response

        return self.get_response(req)



from contextlib import contextmanager

from hirefire_resource.configuration import Configuration


class HireFire:
    configuration = Configuration()

    @classmethod
    @contextmanager
    def configure(cls):
        yield cls.configuration



from hirefire_resource.middleware.asgi import RequestInfo, request


class HireFireMiddleware:
    def __init__(self, app):
        self.original_app = app.asgi_app

    async def __call__(self, scope, receive, send):
        if scope["type"] == "http":
            token = None
            for header_name, header_value in scope["headers"]:
                if header_name.lower() == b"hirefire-token":
                    token = header_value.decode("utf-8")
                    break

            response = await request(
                RequestInfo(
                    path=scope["path"],
                    request_start_time=self.extract_request_start_time(scope),
                    token=token,
                )
            )

            if response:
                await self.send_response(send, *response)
                return

        await self.original_app(scope, receive, send)

    async def send_response(self, send, status, headers, body):
        response_headers = [
            (k.encode("utf-8"), v.encode("utf-8")) for k, v in headers.items()
        ]
        await send(
            {
                "type": "http.response.start",
                "status": status,
                "headers": response_headers,
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": body.encode("utf-8"),
            }
        )

    def extract_request_start_time(self, scope):
        for header_name, header_value in scope["headers"]:
            if header_name.lower() == b"x-request-start":
                return int(header_value.decode("utf-8"))
        return None



import asyncio
import functools
import json
import os
import time
from datetime import datetime, timezone

from celery import Celery
from celery.signals import before_task_publish
from dateutil import parser
from dateutil.parser import parse
from kombu.exceptions import OperationalError

try:
    from amqp.exceptions import ChannelError

    AMQP_AVAILABLE = True
except ImportError:

    class ChannelError(Exception):
        pass

    AMQP_AVAILABLE = False

from hirefire_resource.errors import MissingQueueError


def job_queue_latency(*queues, broker_url=None):
    """
    Calculates the maximum job queue latency across the specified queues using Celery with either
    Redis or RabbitMQ (AMQP) as the broker.

    This function dynamically selects the broker based on the provided broker_url, environment
    variables, or falls back to a default local broker URL. If RabbitMQ (AMQP) is available, it is
    preferred; otherwise, Redis is used.

    Note:
        - Due to Celery's architecture, it is not possible to measure job queue latency with 100%
          accuracy. This function attempts to measure latency as accurately as possible, but there
          are some caveats. See the remaining notes for more details.
        - The `run_at` header is added to each task at publish time using a Celery signal. This
          signal is automatically registered when importing this module, so ensure that every Python
          process that enqueues tasks imports this module.
        - Job queue latency is measured by inspecting the `run_at` header of the next job in the
          queue. For Redis, this works fine, as the queue can be inspected without mutation. For
          RabbitMQ, however, this involves consuming a task, and then rejecting and requeuing it.
          This will occasionally lead to out-of-order execution for certain tasks. However, if
          autoscaling is working effectively, this should not be a significant issue.
        - It is recommended to avoid using the eta and countdown options for tasks in queues that
          are being autoscaled. Tasks with an eta (scheduled tasks) are placed in the same queue as
          regular tasks and maintain the FIFO order, which interferes with job queue latency
          measurement.
        - Failed tasks that are to be retried are published with an eta. While not ideal, occasional
          scheduled tasks, which are typically rare, generally aren't an issue.
        - If you absolutely require the ability to schedule tasks to run in the future, consider
          using a workaround, such as a separate queue for scheduled tasks that forwards tasks ready
          to run to the relevant regular queues. Just remember that the `run_at` header is required.

    Args:
        *queues (str): Names of the queues for latency measurement.
        broker_url (str, optional): The broker URL. Defaults in the following order:
            - Passed argument `broker_url`.
            - Environment variables `AMQP_URL`, `RABBITMQ_URL`, `RABBITMQ_BIGWIG_URL`,
              `CLOUDAMQP_URL`, `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`,
              `OPENREDIS_URL`.
            - "amqp://guest:guest@localhost:5672" if AMQP is available, otherwise
              "redis://localhost:6379/0".

    Returns:
        float: The maximum latency in seconds across the specified queues.

    Raises:
        MissingQueueError: If no queue names are provided.

    Examples:
        >>> job_queue_latency("celery")
        10.172
        >>> job_queue_latency("celery", "mailer")
        22.918
        >>> job_queue_latency("celery", broker_url="amqp://guest:guest@localhost:5672")
        10.172
        >>> job_queue_latency("celery", "mailer", broker_url="redis://localhost:6379/0")
        22.918
    """
    if not queues:
        raise MissingQueueError()

    broker_url = (
        broker_url
        or os.environ.get("AMQP_URL")
        or os.environ.get("RABBITMQ_URL")
        or os.environ.get("RABBITMQ_BIGWIG_URL")
        or os.environ.get("CLOUDAMQP_URL")
        or os.environ.get("REDIS_TLS_URL")
        or os.environ.get("REDIS_URL")
        or os.environ.get("REDISTOGO_URL")
        or os.environ.get("REDISCLOUD_URL")
        or os.environ.get("OPENREDIS_URL")
    )

    if not broker_url:
        if AMQP_AVAILABLE:
            broker_url = "amqp://guest:guest@localhost:5672"
        else:
            broker_url = "redis://localhost:6379/0"

    app = Celery(broker=broker_url)

    try:
        with app.connection_or_acquire() as connection:
            with connection.channel() as channel:
                if hasattr(channel, "_size"):
                    fn = _job_queue_latency_redis
                else:
                    fn = _job_queue_latency_rabbitmq

                return max(fn(channel, queue) for queue in queues)

    except OperationalError:
        return 0


async def async_job_queue_latency(*queues, broker_url=None):
    """
    Asynchronously calculates the maximum job queue latency across the specified queues using Celery
    with either Redis or RabbitMQ (AMQP) as the broker.

    This function is an asynchronous wrapper around the synchronous `job_queue_latency` function. It
    executes the synchronous function in a separate thread using asyncio's event loop and
    `run_in_executor` method. This ensures that the synchronous Celery I/O operations do not block
    the asyncio event loop.

    Note:
        - Due to Celery's architecture, it is not possible to measure job queue latency with 100%
          accuracy. This function attempts to measure latency as accurately as possible, but there
          are some caveats. See the remaining notes for more details.
        - The `run_at` header is added to each task at publish time using a Celery signal. This
          signal is automatically registered when importing this module, so ensure that every Python
          process that enqueues tasks imports this module.
        - Job queue latency is measured by inspecting the `run_at` header of the next job in the
          queue. For Redis, this works fine, as the queue can be inspected without mutation. For
          RabbitMQ, however, this involves consuming a task, and then rejecting and requeuing it.
          This will occasionally lead to out-of-order execution for certain tasks. However, if
          autoscaling is working effectively, this should not be a significant issue.
        - It is recommended to avoid using the eta and countdown options for tasks in queues that
          are being autoscaled. Tasks with an eta (scheduled tasks) are placed in the same queue as
          regular tasks and maintain the FIFO order, which interferes with job queue latency
          measurement.
        - Failed tasks that are to be retried are published with an eta. While not ideal, occasional
          scheduled tasks, which are typically rare, generally aren't an issue.
        - If you absolutely require the ability to schedule tasks to run in the future, consider
          using a workaround, such as a separate queue for scheduled tasks that forwards tasks ready
          to run to the relevant regular queues. Just remember that the `run_at` header is required.

    Args:
        *queues (str): Names of the queues for latency measurement.
        broker_url (str, optional): The broker URL. Defaults in the following order:
            - Passed argument `broker_url`.
            - Environment variables `AMQP_URL`, `RABBITMQ_URL`, `RABBITMQ_BIGWIG_URL`,
              `CLOUDAMQP_URL`, `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`,
              `OPENREDIS_URL`.
            - "amqp://guest:guest@localhost:5672" if AMQP is available, otherwise
              "redis://localhost:6379/0".

    Returns:
        float: The maximum latency in seconds across the specified queues.

    Raises:
        MissingQueueError: If no queue names are provided.

    Examples:
        >>> await async_job_queue_latency("celery")
        10.172
        >>> await async_job_queue_latency("celery", "mailer")
        22.918
        >>> await async_job_queue_latency("celery", broker_url="amqp://guest:guest@localhost:5672")
        10.172
        >>> await async_job_queue_latency("celery", "mailer", broker_url="redis://localhost:6379/0")
        22.918
    """
    loop = asyncio.get_event_loop()
    func = functools.partial(job_queue_latency, *queues, broker_url=broker_url)
    return await loop.run_in_executor(None, func)


def job_queue_size(*queues, broker_url=None):
    """
    Calculates the total job queue size across the specified queues using Celery with either Redis
    or RabbitMQ (AMQP) as the broker.

    This function dynamically selects the broker based on the provided broker_url, environment
    variables, or falls back to a default local broker URL. If RabbitMQ (AMQP) is available, it is
    preferred; otherwise, Redis is used.

    Note:
        - It is recommended to avoid using the eta and countdown options for tasks in queues that
          are being autoscaled. Tasks with an eta (scheduled tasks) are placed in the same queue as
          regular tasks, which interferes with job queue size measurement.
        - Failed tasks that are to be retried are published with an eta. While not ideal, occasional
          scheduled tasks, which are typically rare, generally aren't an issue.
        - If you absolutely require the ability to schedule tasks to run in the future, consider
          using a workaround, such as a separate queue for scheduled tasks that forwards tasks ready
          to run to the relevant regular queues. When using RabbitMQ (AMQP), consider using the
          Delayed Message Plugin.

    Args:
        *queues (str): Names of the queues for size measurement.
        broker_url (str, optional): The broker URL. Defaults in the following order:
            - Passed argument `broker_url`.
            - Environment variables `AMQP_URL`, `RABBITMQ_URL`, `RABBITMQ_BIGWIG_URL`,
              `CLOUDAMQP_URL`, `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`,
              `OPENREDIS_URL`.
            - "amqp://guest:guest@localhost:5672" if AMQP is available, otherwise
              "redis://localhost:6379/0".

    Returns:
        int: The cumulative job queue size across the specified queues.

    Raises:
        MissingQueueError: If no queue names are provided.

    Examples:
        >>> job_queue_size("celery")
        42
        >>> job_queue_size("celery", "mailer")
        85
        >>> job_queue_size("celery", broker_url="amqp://user:password@host:5672")
        42
        >>> job_queue_size("celery", broker_url="redis://localhost:6379/0")
        42
    """
    if not queues:
        raise MissingQueueError()

    broker_url = (
        broker_url
        or os.environ.get("AMQP_URL")
        or os.environ.get("RABBITMQ_URL")
        or os.environ.get("RABBITMQ_BIGWIG_URL")
        or os.environ.get("CLOUDAMQP_URL")
        or os.environ.get("REDIS_TLS_URL")
        or os.environ.get("REDIS_URL")
        or os.environ.get("REDISTOGO_URL")
        or os.environ.get("REDISCLOUD_URL")
        or os.environ.get("OPENREDIS_URL")
    )

    if not broker_url:
        if AMQP_AVAILABLE:
            broker_url = "amqp://guest:guest@localhost:5672"
        else:
            broker_url = "redis://localhost:6379/0"

    app = Celery(broker=broker_url)

    try:
        with app.connection_or_acquire() as connection:
            with connection.channel() as channel:
                worker_task_count = _job_queue_size_worker(app, queues)
                broker_task_count = _job_queue_size_broker(channel, queues)
                return worker_task_count + broker_task_count

    except OperationalError:
        return 0


async def async_job_queue_size(*queues, broker_url=None):
    """
    Asynchronously calculates the total job queue size across the specified queues using Celery with
    either Redis or RabbitMQ (AMQP) as the broker.

    This function is an asynchronous wrapper around the synchronous `job_queue_size` function. It
    executes the synchronous function in a separate thread using asyncio's event loop and
    `run_in_executor` method. This ensures that the synchronous Celery I/O operations do not block
    the asyncio event loop.

    Note:
        - It is recommended to avoid using the eta and countdown options for tasks in queues that
          are being autoscaled. Tasks with an eta (scheduled tasks) are placed in the same queue as
          regular tasks, which interferes with job queue size measurement.
        - Failed tasks that are to be retried are published with an eta. While not ideal, occasional
          scheduled tasks, which are typically rare, generally aren't an issue.
        - If you absolutely require the ability to schedule tasks to run in the future, consider
          using a workaround, such as a separate queue for scheduled tasks that forwards tasks ready
          to run to the relevant regular queues. When using RabbitMQ (AMQP), consider using the
          Delayed Message Plugin.

    Args:
        *queues (str): Names of the queues for size measurement.
        broker_url (str, optional): The broker URL. Defaults in the following order:
            - Passed argument `broker_url`.
            - Environment variables `AMQP_URL`, `RABBITMQ_URL`, `RABBITMQ_BIGWIG_URL`,
              `CLOUDAMQP_URL`, `REDIS_TLS_URL`, `REDIS_URL`, `REDISTOGO_URL`, `REDISCLOUD_URL`,
              `OPENREDIS_URL`.
            - "amqp://guest:guest@localhost:5672" if AMQP is available, otherwise
              "redis://localhost:6379/0".

    Returns:
        int: The cumulative job queue size across the specified queues.

    Raises:
        MissingQueueError: If no queue names are provided.

    Examples:
        >>> await async_job_queue_size("celery")
        42
        >>> await async_job_queue_size("celery", "mailer")
        85
        >>> await async_job_queue_size("celery", broker_url="amqp://user:password@host:5672")
        42
        >>> await async_job_queue_size("celery", broker_url="redis://localhost:6379/0")
        42
    """
    loop = asyncio.get_event_loop()
    func = functools.partial(job_queue_size, *queues, broker_url=broker_url)
    return await loop.run_in_executor(None, func)


@before_task_publish.connect
def run_at_header_signal(
    sender=None, headers=None, body=None, properties=None, **kwargs
):
    headers = headers or {}
    eta = headers.get("eta")

    if eta:
        headers["run_at"] = eta
    else:
        headers["run_at"] = datetime.now(timezone.utc).isoformat()


def _job_queue_latency_redis(channel, queue):
    oldest_job = channel.client.lindex(queue, -1)

    if oldest_job:
        oldest_job = json.loads(oldest_job.decode("utf-8"))
        run_at = oldest_job.get("headers", {}).get("run_at")

        if run_at:
            run_at_time = parse(run_at)
            latency = time.time() - run_at_time.timestamp()
            return max(0, latency)

    return 0


def _job_queue_latency_rabbitmq(channel, queue):
    try:
        message = channel.basic_get(queue)

        if message is None:
            return 0

        run_at = message.headers.get("run_at")

        if run_at:
            run_at_time = parse(run_at)
            latency = time.time() - run_at_time.timestamp()
            result = max(0, latency)
        else:
            result = 0

        channel.basic_reject(message.delivery_tag, requeue=True)

        return result
    except ChannelError:
        return 0


def _job_queue_size_worker(app, queues):
    worker_data = _worker_data(app)
    return sum(worker_data.get(queue, 0) for queue in queues)


def _job_queue_size_broker(channel, queues):
    if hasattr(channel, "_size"):
        fn = _job_queue_size_redis
    else:
        fn = _job_queue_size_rabbitmq

    return sum(fn(channel, queue) for queue in queues)


def _job_queue_size_redis(channel, queue):
    return channel.client.llen(queue)


def _job_queue_size_rabbitmq(channel, queue):
    try:
        return channel.queue_declare(queue=queue, passive=True).message_count
    except ChannelError:
        return 0


_worker_data_cache_enabled = True
_worker_data_cache_value = None
_worker_data_cache_time = time.time() - (5 + 1)


# Used to disable the worker data cache for testing
def _cache_worker_data(enabled):
    global _worker_data_cache_enabled
    _worker_data_cache_enabled = enabled


def _worker_data(app):
    global _worker_data_cache_value, _worker_data_cache_time

    if not _worker_data_cache_enabled or (_worker_data_cache_time + 5) < time.time():
        i = app.control.inspect()
        now = time.time()
        queue_info = {}

        for collection in [i.active(), i.reserved(), i.scheduled()]:
            if collection is not None:
                for worker, tasks in collection.items():
                    for task in tasks:
                        task_info = task

                        if task.get("eta"):
                            eta_string = task.get("eta")
                            eta_datetime = parser.parse(eta_string)
                            eta_timestamp = eta_datetime.timestamp()

                            if now < eta_timestamp:
                                continue

                            task_info = task["request"]

                        queue = task_info["delivery_info"]["routing_key"]

                        if queue not in queue_info:
                            queue_info[queue] = 0

                        queue_info[queue] += 1

        _worker_data_cache_value = queue_info
        _worker_data_cache_time = time.time()

    return _worker_data_cache_value



import os
import time

from hirefire_resource import HireFire


class RequestInfo:
    def __init__(self, path, request_start_time=None, token=None):
        self.path = path

        if request_start_time:
            self.request_start_time = int(request_start_time)
        else:
            self.request_start_time = None

        self.token = token


def matches_hirefire_path(request_info):
    return (
        os.environ.get("HIREFIRE_TOKEN")
        and request_info.path == "/hirefire"
        and request_info.token == os.environ.get("HIREFIRE_TOKEN")
    )


def matches_info_path(request_info):
    return (
        os.environ.get("HIREFIRE_TOKEN")
        and request_info.path == f"/hirefire/{os.environ.get('HIREFIRE_TOKEN')}/info"
    )


def process_request_queue_time(request_info):
    if not (
        os.environ.get("HIREFIRE_TOKEN")
        and HireFire.configuration.web
        and request_info.request_start_time
    ):
        return

    request_queue_time = calculate_request_queue_time(request_info)

    HireFire.configuration.web.start_dispatcher()
    HireFire.configuration.web.add_to_buffer(request_queue_time)


def calculate_request_queue_time(request_info):
    return max(int(time.time() * 1000) - request_info.request_start_time, 0)



import re


class InvalidDynoNameError(Exception):
    pass


class MissingDynoProcError(Exception):
    pass


class Worker:
    PROCESS_NAME_PATTERN = r"^[a-zA-Z][a-zA-Z0-9_-]{0,29}$"

    def __init__(self, name, proc=None):
        self._validate(name, proc)
        self.name = name
        self._proc = proc

    def value(self):
        return self._proc()

    def _validate(self, name, proc):
        if not re.match(self.PROCESS_NAME_PATTERN, name or ""):
            raise InvalidDynoNameError(
                f"Invalid name for Worker({name}, proc). "
                "Ensure it matches the Procfile process name (i.e. web, worker)."
            )

        if proc is None:
            raise MissingDynoProcError(
                f"Missing proc for Worker({name}, proc). "
                "Ensure that you provide a proc that returns the job queue metric."
            )



import json

from hirefire_resource import HireFire
from hirefire_resource.middleware import (  # noqa
    RequestInfo,
    matches_hirefire_path,
    matches_info_path,
    process_request_queue_time,
)
from hirefire_resource.version import VERSION


def request(request_info):
    process_request_queue_time(request_info)

    if matches_hirefire_path(request_info) or matches_info_path(request_info):
        return construct_info_response()


def construct_info_response():
    headers = {
        "Content-Type": "application/json",
        "Cache-Control": "must-revalidate, private, max-age=0",
        "HireFire-Resource": f"Python-{VERSION}",
    }
    body = json.dumps(collect_workers_data())
    return 200, headers, [body]


def collect_workers_data():
    return [
        {"name": worker.name, "value": worker.value()}
        for worker in HireFire.configuration.workers
    ]
